/*
 * cal.c
 *
 * CPU Abstraction Layer implimentation.
 */

/* $Id: cal.c,v 1.19 1999/12/04 04:59:44 nyef Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include "cal.h"

#ifdef CPU_MARAT6502
#include "m6502.h"
#endif

#ifdef CPU_MARATZ80
#include "mz80.h"
#endif

#ifdef CPU_JUNK68K
#include "emu68k.h"
#endif

#ifdef CPU_MARAT6280
#include "m6280.h"
#endif

#ifdef CPU_MARAT6502
void cal_marat6502_init(cal_cpu *cpu);
void cal_marat6502_reset(cal_cpu cpu);
void cal_marat6502_run(cal_cpu cpu);
void cal_marat6502_runfor(cal_cpu cpu, int cycles);
void cal_marat6502_irq(cal_cpu cpu, int irqno);
void cal_marat6502_nmi(cal_cpu cpu);
void cal_marat6502_setzpage(cal_cpu cpu, void *page0);
int cal_marat6502_timeleft(cal_cpu cpu);
void cal_marat6502_setmmu0(cal_cpu cpu, ranged_mmu *mmu);
void cal_marat6502_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl);
void cal_marat6502_setiou(cal_cpu cpu, memread8_t ioread, memwrite8_t iowrite);
#endif

#ifdef CPU_MARAT6280
void cal_marat6280_init(cal_cpu *cpu);
void cal_marat6280_reset(cal_cpu cpu);
void cal_marat6280_run(cal_cpu cpu);
void cal_marat6280_runfor(cal_cpu cpu, int cycles);
void cal_marat6280_irq(cal_cpu cpu, int irqno);
void cal_marat6280_nmi(cal_cpu cpu);
void cal_marat6280_setzpage(cal_cpu cpu, void *page0);
int cal_marat6280_timeleft(cal_cpu cpu);
void cal_marat6280_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl);
void cal_marat6280_setiou(cal_cpu cpu, memread8_t ioread, memwrite8_t iowrite);
#endif

#ifdef CPU_MARATZ80
void cal_maratz80_init(cal_cpu *cpu);
void cal_maratz80_reset(cal_cpu cpu);
void cal_maratz80_run(cal_cpu cpu);
void cal_maratz80_runfor(cal_cpu cpu, int cycles);
void cal_maratz80_irq(cal_cpu cpu, int irqno);
void cal_maratz80_nmi(cal_cpu cpu);
void cal_maratz80_setzpage(cal_cpu cpu, void *page0);
int cal_maratz80_timeleft(cal_cpu cpu);
void cal_maratz80_setmmu0(cal_cpu cpu, ranged_mmu *mmu);
void cal_maratz80_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl);
void cal_maratz80_setiou(cal_cpu cpu, memread8_t ioread, memwrite8_t iowrite);
#endif

#ifdef CPU_JUNK68K
void cal_junk68k_init(cal_cpu *cpu);
void cal_junk68k_reset(cal_cpu cpu);
void cal_junk68k_run(cal_cpu cpu);
void cal_junk68k_runfor(cal_cpu cpu, int cycles);
void cal_junk68k_irq(cal_cpu cpu, int irqno);
void cal_junk68k_nmi(cal_cpu cpu);
void cal_junk68k_setzpage(cal_cpu cpu, void *page0);
int cal_junk68k_timeleft(cal_cpu cpu);
void cal_junk68k_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl);
void cal_junk68k_setmmu16(cal_cpu cpu, int shift, int mask, memread16_t *rtbl, memwrite16_t *wtbl);
#endif

#ifdef CPU_TORR68K
void cal_torr68k_init(cal_cpu *cpu);
#endif

typedef void (*cal_cpuinit_t)(cal_cpu *cpu);

struct cal_cputype {
    int type;
    cal_cpuinit_t init;
} cal_cputypes[] = {
#ifdef CPU_MARAT6502
    {CPUT_MARAT6502, cal_marat6502_init},
#endif
#ifdef CPU_MARATZ80
    {CPUT_MARATZ80, cal_maratz80_init},
#endif
#ifdef CPU_JUNK68K
    {CPUT_JUNK68K, cal_junk68k_init},
#endif
#ifdef CPU_TORR68K
    {CPUT_TORR68K, cal_torr68k_init},
#endif
#ifdef CPU_MARAT6280
    {CPUT_MARAT6280, cal_marat6280_init},
#endif
    {0, NULL}, /* must be last entry */
};

cal_cpu cal_create(int type)
{
    cal_cpu retval;
    struct cal_cputype *cputype;

    retval = calloc(1, sizeof(struct cal_cpu));

    if (retval) {
	retval->type = type;

	for (cputype = cal_cputypes; cputype->init; cputype++) {
	    if (cputype->type == type) break;
	}
	
	if (cputype->init) {
	    cputype->init(&retval);
	} else {
	    printf("Unknown CPU type.\n");
	    free(retval);
	    retval = NULL;
	}
    } else {
	printf("Insufficient memory to create CPU.\n");
    }

    return retval;
}

#ifdef CPU_MARAT6502

void cal_marat6502_init(cal_cpu *cpu)
{
    (*cpu)->data.d_marat6502 = calloc(1, sizeof(M6502));
    if (!(*cpu)->data.d_marat6502) {
	printf("Insufficient memory to create CPU.\n");
	free(*cpu);
	*cpu = NULL;
	return;
    }
    (*cpu)->reset = cal_marat6502_reset;
    (*cpu)->run = cal_marat6502_run;
    (*cpu)->runfor = cal_marat6502_runfor;
    (*cpu)->irq = cal_marat6502_irq;
    (*cpu)->nmi = cal_marat6502_nmi;
    (*cpu)->setzpage = cal_marat6502_setzpage;
    (*cpu)->timeleft = cal_marat6502_timeleft;
    (*cpu)->setmmu0 = cal_marat6502_setmmu0;
    (*cpu)->setmmu8 = cal_marat6502_setmmu8;
    (*cpu)->setiou = cal_marat6502_setiou;
    (*cpu)->data.d_marat6502->User = *cpu;
}

void cal_marat6502_reset(cal_cpu cpu)
{
    Reset6502(cpu->data.d_marat6502);
}

void cal_marat6502_run(cal_cpu cpu)
{
    Run6502(cpu->data.d_marat6502);
}

void cal_marat6502_runfor(cal_cpu cpu, int cycles)
{
    cpu->data.d_marat6502->IPeriod = cycles;
}

void cal_marat6502_irq(cal_cpu cpu, int irqno)
{
    cpu->data.d_marat6502->IRequest = INT_IRQ;
}

void cal_marat6502_nmi(cal_cpu cpu)
{
    cpu->data.d_marat6502->IRequest = INT_NMI;
}

void cal_marat6502_setzpage(cal_cpu cpu, void *page0)
{
}

int cal_marat6502_timeleft(cal_cpu cpu)
{
    return cpu->data.d_marat6502->ICount;
}

void cal_marat6502_setmmu0(cal_cpu cpu, ranged_mmu *mmu)
{
    cpu->data.d_marat6502->mmu = mmu;
}

void cal_marat6502_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl)
{
    cpu->data.d_marat6502->memshift = shift;
    cpu->data.d_marat6502->memmask = mask;
    cpu->data.d_marat6502->readtable = rtbl;
    cpu->data.d_marat6502->writetable = wtbl;
}

void cal_marat6502_setiou(cal_cpu cpu, memread8_t ioread, memwrite8_t iowrite)
{
}

#endif


#ifdef CPU_MARATZ80

void cal_maratz80_init(cal_cpu *cpu)
{
    (*cpu)->data.d_maratz80 = calloc(1, sizeof(Z80));
    if (!(*cpu)->data.d_maratz80) {
	printf("Insufficient memory to create CPU.\n");
	free(*cpu);
	*cpu = NULL;
	return;
    }
    (*cpu)->reset = cal_maratz80_reset;
    (*cpu)->run = cal_maratz80_run;
    (*cpu)->runfor = cal_maratz80_runfor;
    (*cpu)->irq = cal_maratz80_irq;
    (*cpu)->nmi = cal_maratz80_nmi;
    (*cpu)->setzpage = cal_maratz80_setzpage;
    (*cpu)->timeleft = cal_maratz80_timeleft;
    (*cpu)->setmmu0 = cal_maratz80_setmmu0;
    (*cpu)->setmmu8 = cal_maratz80_setmmu8;
    (*cpu)->setiou = cal_maratz80_setiou;
    (*cpu)->data.d_maratz80->User = *cpu;
}

void cal_maratz80_reset(cal_cpu cpu)
{
    ResetZ80(cpu->data.d_maratz80);
}

void cal_maratz80_run(cal_cpu cpu)
{
    RunZ80(cpu->data.d_maratz80);
}

void cal_maratz80_runfor(cal_cpu cpu, int cycles)
{
    cpu->data.d_maratz80->IPeriod = cycles;
}

void cal_maratz80_irq(cal_cpu cpu, int irqno)
{
    cpu->data.d_maratz80->IRequest = INT_IRQ;
}

void cal_maratz80_nmi(cal_cpu cpu)
{
    cpu->data.d_maratz80->IRequest = INT_NMI;
}

void cal_maratz80_setzpage(cal_cpu cpu, void *page0)
{
}

int cal_maratz80_timeleft(cal_cpu cpu)
{
    return cpu->data.d_maratz80->ICount;
}

void cal_maratz80_setmmu0(cal_cpu cpu, ranged_mmu *mmu)
{
    cpu->data.d_maratz80->mmu = mmu;
}

void cal_maratz80_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl)
{
/*     cpu->data.d_maratz80->memshift = shift; */
/*     cpu->data.d_maratz80->memmask = mask; */
/*     cpu->data.d_maratz80->readtable = rtbl; */
/*     cpu->data.d_maratz80->writetable = wtbl; */
    cpu->data.d_maratz80->readfunc = *rtbl;
    cpu->data.d_maratz80->writefunc = *wtbl;
}

void cal_maratz80_setiou(cal_cpu cpu, memread8_t ioread, memwrite8_t iowrite)
{
    cpu->data.d_maratz80->io_read = ioread;
    cpu->data.d_maratz80->io_write = iowrite;
}

#endif


#ifdef CPU_JUNK68K

void cal_junk68k_init(cal_cpu *cpu)
{
    (*cpu)->data.d_junk68k = calloc(1, sizeof(struct emu68k_context));
    if (!(*cpu)->data.d_junk68k) {
	printf("Insufficient memory to create CPU.\n");
	free(*cpu);
	*cpu = NULL;
	return;
    }
    (*cpu)->reset = cal_junk68k_reset;
    (*cpu)->run = cal_junk68k_run;
    (*cpu)->runfor = cal_junk68k_runfor;
    (*cpu)->irq = cal_junk68k_irq;
    (*cpu)->nmi = cal_junk68k_nmi;
    (*cpu)->setzpage = cal_junk68k_setzpage;
    (*cpu)->timeleft = cal_junk68k_timeleft;
    (*cpu)->setmmu8 = cal_junk68k_setmmu8;
    (*cpu)->setmmu16 = cal_junk68k_setmmu16;
    (*cpu)->data.d_junk68k->cpu = *cpu;
/*     emuz80_init((*cpu)->data.d_junk68k); */
}

void cal_junk68k_reset(cal_cpu cpu)
{
    emu68k_reset(cpu->data.d_junk68k);
}

void cal_junk68k_run(cal_cpu cpu)
{
    emu68k_step(cpu->data.d_junk68k);
}

void cal_junk68k_runfor(cal_cpu cpu, int cycles)
{
/*     cpu->data.d_emuz80->cycles_left += cycles; */
}

void cal_junk68k_irq(cal_cpu cpu, int irqno)
{
/*     emuz80_IRQ(cpu->data.d_emuz80); */
}

void cal_junk68k_nmi(cal_cpu cpu)
{
/*     emuz80_NMI(cpu->data.d_emuz80); */
}

void cal_junk68k_setzpage(cal_cpu cpu, void *page0)
{
}

int cal_junk68k_timeleft(cal_cpu cpu)
{
/*     return cpu->data.d_emuz80->cycles_left; */
    return 0;
}

void cal_junk68k_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl)
{
    cpu->data.d_junk68k->memshift = shift;
    cpu->data.d_junk68k->memmask = mask;
    cpu->data.d_junk68k->read8table = rtbl;
    cpu->data.d_junk68k->write8table = wtbl;
}

void cal_junk68k_setmmu16(cal_cpu cpu, int shift, int mask, memread16_t *rtbl, memwrite16_t *wtbl)
{
    cpu->data.d_junk68k->read16table = rtbl;
    cpu->data.d_junk68k->write16table = wtbl;
}

#endif


#ifdef CPU_MARAT6280

void cal_marat6280_init(cal_cpu *cpu)
{
    (*cpu)->data.d_marat6280 = calloc(1, sizeof(M6280));
    if (!(*cpu)->data.d_marat6280) {
	printf("Insufficient memory to create CPU.\n");
	free(*cpu);
	*cpu = NULL;
	return;
    }
    (*cpu)->reset = cal_marat6280_reset;
    (*cpu)->run = cal_marat6280_run;
    (*cpu)->runfor = cal_marat6280_runfor;
    (*cpu)->irq = cal_marat6280_irq;
    (*cpu)->nmi = cal_marat6280_nmi;
    (*cpu)->setzpage = cal_marat6280_setzpage;
    (*cpu)->timeleft = cal_marat6280_timeleft;
    (*cpu)->setmmu8 = cal_marat6280_setmmu8;
    (*cpu)->setiou = cal_marat6280_setiou;
    (*cpu)->data.d_marat6280->User = *cpu;
}

void cal_marat6280_reset(cal_cpu cpu)
{
    Reset6280(cpu->data.d_marat6280);
}

void cal_marat6280_run(cal_cpu cpu)
{
    Run6280(cpu->data.d_marat6280);
}

void cal_marat6280_runfor(cal_cpu cpu, int cycles)
{
    cpu->data.d_marat6280->IPeriod = cycles;
}

void cal_marat6280_irq(cal_cpu cpu, int irqno)
{
    cpu->data.d_marat6280->IRequest = INT_IRQ;
}

void cal_marat6280_nmi(cal_cpu cpu)
{
    cpu->data.d_marat6280->IRequest = INT_NMI;
}

void cal_marat6280_setzpage(cal_cpu cpu, void *page0)
{
    cpu->data.d_marat6280->page_0 = page0;
}

int cal_marat6280_timeleft(cal_cpu cpu)
{
    return cpu->data.d_marat6280->ICount;
}

void cal_marat6280_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl)
{
    cpu->data.d_marat6280->memshift = shift;
    cpu->data.d_marat6280->memmask = mask;
    cpu->data.d_marat6280->readtable = rtbl;
    cpu->data.d_marat6280->writetable = wtbl;
}

void cal_marat6280_setiou(cal_cpu cpu, memread8_t ioread, memwrite8_t iowrite)
{
}

#endif

/*
 * $Log: cal.c,v $
 * Revision 1.19  1999/12/04 04:59:44  nyef
 * cleared out all code relating to the ASM CPU cores
 *
 * Revision 1.18  1999/11/21 04:39:51  nyef
 * added implementation of ranged mmu for marat 6502 core
 *
 * Revision 1.17  1999/11/21 04:26:29  nyef
 * changed cal_create() to use loop and collection idiom
 *
 * Revision 1.16  1999/11/08 01:31:42  nyef
 * added implementation of ranged mmu for marat z80 core
 *
 * Revision 1.15  1999/08/29 13:05:56  nyef
 * updated to new IRQ interface
 *
 * Revision 1.14  1999/08/15 02:28:36  nyef
 * added TORR68K support
 *
 * Revision 1.13  1999/08/02 00:16:30  nyef
 * added marat6280 support
 * added conditional compilation of junk68k support
 *
 * Revision 1.12  1999/06/06 17:50:22  nyef
 * changed maratz80 memory access interface to be faster
 *
 * Revision 1.11  1999/04/17 20:06:10  nyef
 * added 68k support
 *
 * Revision 1.10  1999/01/24 15:58:33  nyef
 * changed all malloc(x) to calloc(1, x).
 *
 * Revision 1.9  1999/01/17 02:10:27  nyef
 * fixed support for irqs in the emu6280 core
 *
 * Revision 1.8  1999/01/13 03:52:23  nyef
 * continued preliminary work on 6280 support.
 *
 * Revision 1.7  1999/01/03 02:25:32  nyef
 * added maratz80 support.
 *
 * Revision 1.6  1999/01/02 06:37:43  nyef
 * fixed stupid mistake with z80 setiou.
 *
 * Revision 1.5  1999/01/02 06:35:40  nyef
 * added io interface for z80 support.
 *
 * Revision 1.4  1999/01/01 07:00:59  nyef
 * added emuz80 support
 *
 * Revision 1.3  1998/12/12 21:34:37  nyef
 * brought remaining cpu interfaces up to spec
 *
 * Revision 1.2  1998/12/12 20:34:22  nyef
 * brought marat6502 interface up to spec.
 *
 * Revision 1.1  1998/11/29 18:28:44  nyef
 * Initial revision
 *
 */
