/*
 * cal.h
 *
 * CPU Abstraction Layer interface.
 */

/* $Id: cal.h,v 1.12 1999/11/08 01:31:11 nyef Exp $ */

#ifndef __CAL_H__
#define __CAL_H__

#include "cpudefs.h"

#define CPUT_NONE      0
#define CPUT_MARAT6502 1
#define CPUT_EMU6502   2
#define CPUT_EMU6502A  3
#define CPUT_EMU65C02  4
#define CPUT_EMU6280   5
#define CPUT_MARATZ80  6
#define CPUT_EMUZ80    7
#define CPUT_JUNK68K   8
#define CPUT_MARAT6280 9
#define CPUT_TORR68K   10

typedef struct cal_cpu *cal_cpu;

typedef unsigned char (*memread8_t)(cal_cpu, unsigned long);
typedef void (*memwrite8_t)(cal_cpu, unsigned long, unsigned char);
typedef unsigned short (*memread16_t)(cal_cpu, unsigned long);
typedef void (*memwrite16_t)(cal_cpu, unsigned long, unsigned short);
typedef unsigned long (*memread32_t)(cal_cpu, unsigned long);
typedef void (*memwrite32_t)(cal_cpu, unsigned long, unsigned long);

typedef struct {
    unsigned long start;
    unsigned long end;
    unsigned char *data;
    unsigned short flags;
    unsigned short tag;
} ranged_mmu;

typedef void (*cpureset_t)(cal_cpu);
typedef void (*cpurun_t)(cal_cpu);
typedef void (*cpurunfor_t)(cal_cpu, int);
typedef void (*cpuirq_t)(cal_cpu, int);
typedef void (*cpunmi_t)(cal_cpu);
typedef void (*cpusetzpage_t)(cal_cpu, void *);
typedef int (*cputimeleft_t)(cal_cpu);
typedef void (*cpusetmmu0_t)(cal_cpu, ranged_mmu *);
typedef void (*cpusetmmu8_t)(cal_cpu, int, int, memread8_t *, memwrite8_t *);
typedef void (*cpusetmmu16_t)(cal_cpu, int, int, memread16_t *, memwrite16_t *);
typedef void (*cpusetmmu32_t)(cal_cpu, int, int, memread32_t *, memwrite32_t *);
typedef void (*cpusetiou_t)(cal_cpu, memread8_t, memwrite8_t);

struct cal_cpu {
    int type; /* CPUT_xxx */
    void *userdata;
    cpureset_t reset;
    cpurun_t run;
    cpurunfor_t runfor;
    cpuirq_t irq;
    cpunmi_t nmi;
    cpusetzpage_t setzpage;
    cputimeleft_t timeleft;
    cpusetmmu0_t setmmu0;
    cpusetmmu8_t setmmu8;
    cpusetmmu16_t setmmu16;
    cpusetmmu32_t setmmu32;
    cpusetiou_t setiou;
    union {
#ifdef CPU_MARAT6502
	struct M6502 *d_marat6502;
#endif
#ifdef CPU_EMU6502
	struct emu6502_context *d_emu6502;
#endif
#ifdef CPU_EMU6502A
	struct emu6502_context *d_emu6502a;
#endif
#ifdef CPU_EMU65C02
	struct emu6502_context *d_emu65c02;
#endif
#ifdef CPU_EMU6280
	struct emu6280_context *d_emu6280;
#endif
#ifdef CPU_MARATZ80
	struct Z80 *d_maratz80;
#endif
#ifdef CPU_EMUZ80
	struct emuz80_context *d_emuz80;
#endif
#if defined(CPU_JUNK68K) || defined(CPU_TORR68K)
	struct emu68k_context *d_junk68k;
#endif
#ifdef CPU_MARAT6280
	struct M6280 *d_marat6280;
#endif
    } data;
};

cal_cpu cal_create(int type);

#endif /* __CAL_H__ */

/*
 * $Log: cal.h,v $
 * Revision 1.12  1999/11/08 01:31:11  nyef
 * added preliminary definition and interface for ranged mmu
 *
 * Revision 1.11  1999/08/29 13:06:00  nyef
 * updated to new IRQ interface
 *
 * Revision 1.10  1999/08/15 02:28:21  nyef
 * added TORR68K support
 *
 * Revision 1.9  1999/08/02 00:17:03  nyef
 * added marat6280 support
 * added conditional compilation of junk68k support
 *
 * Revision 1.8  1999/04/17 20:05:57  nyef
 * changed interface to allow 68k support.
 * added 68k support.
 *
 * Revision 1.7  1999/01/26 03:06:08  nyef
 * expanded "address bus" to 32 bits.
 *
 * Revision 1.6  1999/01/09 19:47:47  nyef
 * added 'void *userdata' to struct cal_cpu.
 *
 * Revision 1.5  1999/01/03 02:25:40  nyef
 * added maratz80 support.
 *
 * Revision 1.4  1999/01/02 06:35:46  nyef
 * added io interface for z80 support.
 *
 * Revision 1.3  1999/01/01 07:00:46  nyef
 * added emuz80 support
 *
 * Revision 1.2  1998/12/12 20:35:13  nyef
 * fixed marat6502 support.
 *
 * Revision 1.1  1998/11/29 18:28:34  nyef
 * Initial revision
 *
 */
