/*
 * nes.h
 *
 * NES specific header.
 */

/* $Id: nes.h,v 1.5 1999/11/20 05:33:35 nyef Exp $ */

#ifndef __NES_H__
#define __NES_H__

#include "tool.h"

extern unsigned char *PPU_memory; /* PPU Ram */

extern unsigned char *bank_0;     /* System Ram */

extern unsigned char *bank_6;     /* Cart Ram */
extern unsigned char *bank_8;     /* Rom Page */
extern unsigned char *bank_A;     /* Rom Page */
extern unsigned char *bank_C;     /* Rom Page */
extern unsigned char *bank_E;     /* Rom Page */

extern unsigned char bank8_reg;
extern unsigned char bankA_reg;
extern unsigned char bankC_reg;
extern unsigned char bankE_reg;

extern memread8_t nes_bank4_read_hook;
extern memread8_t nes_bank6_read_hook;
extern memwrite8_t nes_bank4_write_hook;
extern memwrite8_t nes_bank6_write_hook;

typedef struct nes_rom {
    rom_file file;
    unsigned char *prg_data;
    unsigned char *chr_data;
    int prg_size;
    int chr_size;
    int mapper;
} *nes_rom;

#define cart_rom_size(image) (read_byte(image, 4))
#define cart_vrom_size(image) (read_byte(image, 5))
#define cart_mapper(image) (read_byte(image, 6) >> 4)
#define cart_mapper_ext(image) (read_byte(image, 7) & 0xf0)
#define cart_ext_mapper(image) (cart_mapper_ext(image) | cart_mapper(image))
#define cart_is_ext_mapper(image) (read_long(image, 12) == 0)

void sync_banks(void);

#endif /* __NES_H__ */

/*
 * $Log: nes.h,v $
 * Revision 1.5  1999/11/20 05:33:35  nyef
 * fixed to work with new rom loading interface
 * rebuild rom file handling
 *
 * Revision 1.4  1999/04/17 20:08:57  nyef
 * fixed for new version of the CAL.
 *
 * Revision 1.3  1999/02/19 02:09:41  nyef
 * added hooks for reading and writing banks 4 and 6
 *
 * Revision 1.2  1998/12/31 03:53:14  nyef
 * added support for 8-bit mapper ids.
 *
 * Revision 1.1  1998/07/11 22:18:06  nyef
 * Initial revision
 *
 */
