/*
 * tool.c
 *
 * a program to aide in hacking FF3j
 */

/* $Id: tool.c,v 1.23 2000/01/01 04:12:05 nyef Exp $ */

#include <stdio.h>
#include <stdlib.h>
/* #include <unistd.h> */
/* #include <fcntl.h> */
/* #include <sys/stat.h> */
#include <ctype.h>
#include <string.h>

#include "machdep.h"
#include "ui.h"
#include "video.h"
#include "tool.h"
#include "system.h"

unsigned long system_flags;

#ifdef USE_OLD_GUESS_SYSTEM
static const unsigned char gb_sig[0x30] = {
    0xCE, 0xED, 0x66, 0x66, 0xCC, 0x0D, 0x00, 0x0B,
    0x03, 0x73, 0x00, 0x83, 0x00, 0x0C, 0x00, 0x0D,
    0x00, 0x08, 0x11, 0x1F, 0x88, 0x89, 0x00, 0x0E,
    0xDC, 0xCC, 0x6E, 0xE6, 0xDD, 0xDD, 0xD9, 0x99,
    0xBB, 0xBB, 0x67, 0x63, 0x6E, 0x0E, 0xEC, 0xCC,
    0xDD, 0xDC, 0x99, 0x9F, 0xBB, 0xB9, 0x33, 0x3E
};
#endif

rom_file read_romimage(char *filename)
{
    FILE *file;
    int numread;
    int error;
    rom_file retval;

    file = NULL;
    
    retval = malloc(sizeof(struct rom_file));

    if (retval) {
	retval->filename = NULL;
	retval->data = NULL;
	error = 0;
    } else {
	error = 1;
    }
    
    if (!error) {
	file = fopen(filename, "rb");
	error = !file;
    }
  
    if (!error) {
	/* FIXME: ignoring the return value from fseek() isn't good */
	fseek(file, 0, SEEK_END);
	retval->size = ftell(file);
	fseek(file, 0, SEEK_SET);
	
	retval->filename = strdup(filename);
	retval->data = malloc(retval->size);
	error = !(retval->data && retval->filename);
    }

    if (!error) {
	numread = fread(retval->data, 1, retval->size, file);
	error = (numread != retval->size);
    }

    if (error && retval) {
	if (retval->filename) {
	    free(retval->filename);
	}
	if (retval->data) {
	    free(retval->data);
	}
	free(retval);
	retval = NULL;
    }

    if (file) {
	fclose(file);
    }
    
    return retval;
}

#ifdef MSB_FIRST
#define BATTERY_EXTENSION 0x2e736176 /* ".sav" */
#endif
#ifdef LSB_FIRST
#define BATTERY_EXTENSION 0x7661732e /* ".sav" */
#endif

rom_file init_battery_file(rom_file romfile, int size)
{
    FILE *file;
    rom_file retval;

    file = NULL;
    
    retval = calloc(1, sizeof(struct rom_file));

    if (!retval) {
	return NULL;
    }
    
    /* FIXME: This is Just Plain Wrong */
    retval->filename = strdup(romfile->filename);
    *((unsigned long *) &retval->filename[strlen(retval->filename) - 4]) = BATTERY_EXTENSION;
    
    retval->data = malloc(size);
    retval->size = size;
    
    file = fopen(retval->filename, "rb");

    if (!file) {
	/* No file? just return an empty buffer */
	return retval;
    }

    /* FIXME: error handling? */
    fread(retval->data, 1, retval->size, file);

    fclose(file);
    
    return retval;
}

void save_battery_file(rom_file batteryfile)
{
    FILE *batfile;
    
    deb_printf("attempting to save battery file.\n");
    
    /* FIXME: error checking might be nice */
    
    batfile = fopen(batteryfile->filename, "wb");
    
    if (batfile != NULL) {
	fwrite(batteryfile->data, batteryfile->size, 1, batfile);
	fclose(batfile);
    } else {
	deb_printf("could not open battery file.\n");
    }
}

#ifdef USE_OLD_GUESS_SYSTEM
int guess_system(rom_file romfile)
{
    unsigned char *romimage;
    int romsize;
    int system_type;
    
    romimage = romfile->data;
    romsize = romfile->size;
    system_type = ST_NONE;
    
    if (!memcmp(romimage, "NES\x1a", 4)) {
	deb_printf("file is NES image.\n");
	system_type = ST_NES;
    } else if (!memcmp(romimage, "FDS\x1a", 4)) {
	deb_printf("file is Famicom Disk System image.\n");
	system_type = ST_NES;
    } else if ((((unsigned short *)romimage)[0] == 0x55aa) ||
	       (((unsigned short *)romimage)[0] == 0xaa55)) {
	deb_printf("file is probably ColecoVision image.\n");
	system_type = ST_COLECO;
    } else if (((romsize >= 0x8000) &&
	       (!memcmp(romimage + 0x7ff0, "TMR SEGA", 8))) ||
	       ((romsize >= 0x8200) &&
	       (!memcmp(romimage + 0x81f0, "TMR SEGA", 8)))) {
	deb_printf("file is SMS/GG image.\n");
	system_type = ST_MASTERSYS;
    } else if (!memcmp(romimage + 0x104, gb_sig, 0x30)) {
	deb_printf("file is Gameboy image.\n");
	system_type = ST_GAMEBOY;
    } else if ((romsize % 0x2000) == 0) {
	deb_printf("file is a multiple of 0x2000 bytes data.\n");
	if (!memcmp(&romimage[0x100], "SEGA ", 5)) {
	    deb_printf("file is Genesis image.\n");
	    system_type = ST_GENESIS;
	} else {
	    deb_printf("assuming file is PC-Engine image.\n");
	    system_type = ST_PCENGINE;
	}
    } else if ((romsize % 0x2000) == 0x200) {
	deb_printf("file is a multiple of 0x2000 bytes data with 512-byte header.\n");
	if ((!memcmp(&romimage[0x2280], "SG ", 3)) &&
	    (!memcmp(&romimage[0x0280], "EA", 2))) {
	    deb_printf("file is Genesis image.\n");
	    system_type = ST_GENESIS;
	} else {
	    deb_printf("assuming file is PC-Engine image.\n");
	    system_type = ST_PCENGINE;
	}
    }

    return system_type;
}
#else
struct {
    int type;
    char *file_suffix;
} system_detect[] = {
    {ST_NES,       ".nes"},
    {ST_NES,       ".fds"},
    {ST_MASTERSYS, ".sms"},
    {ST_GAMEGEAR,  ".gg"},
    {ST_COLECO,    ".col"},
#if 0
    {ST_GAMEBOY,   ".gb"},
#endif
    {ST_GENESIS,   ".smd"},
    {ST_GENESIS,   ".bin"},
    {ST_PCENGINE,  ".pce"},
    {ST_SG1000,    ".sg"},
    {ST_NONE, NULL}, /* Must be last entry */
};

int guess_system(rom_file romfile)
{
    int i;
    int len_filename;
    int len_suffix;
    int string_offset;

    len_filename = strlen(romfile->filename);

    for (i = 0; system_detect[i].type != ST_NONE; i++) {
	len_suffix = strlen(system_detect[i].file_suffix);
	string_offset = len_filename - len_suffix;
	if (string_offset < 0) {
	    continue;
	}
	if (!strcasecmp(romfile->filename + string_offset, system_detect[i].file_suffix)) {
	    return system_detect[i].type;
	}
    }
    return ST_NONE;
}
#endif

typedef void (*sysrun_t)(rom_file romfile);

struct {
    int type;
    sysrun_t activate;
} systems[] = {
    {ST_NES,       nes_run},
    {ST_PCENGINE,  pce_run},
    {ST_MASTERSYS, sms_run},
    {ST_GAMEGEAR,  gg_run},
    {ST_GENESIS,   genesis_run},
    {ST_COLECO,    cv_run},
    {ST_SG1000,    sg1k_run},
    {ST_APPLE2,    apple2_run},
    {ST_NONE,      NULL}, /* Nust be last entry */
};

void activate_system(int system_type, rom_file romfile)
{
    int i;

    for (i = 0; systems[i].type != ST_NONE; i++) {
	if (systems[i].type == system_type) {
	    systems[i].activate(romfile);
	    return;
	}
    }

    deb_printf("unknown system type.\n");
}

/*
 * $Log: tool.c,v $
 * Revision 1.23  2000/01/01 04:12:05  nyef
 * added activation entry for the Apple ][
 *
 * Revision 1.22  1999/12/07 00:52:58  nyef
 * made the cheap hack used for the battery filename endian-independant
 *
 * Revision 1.21  1999/12/06 00:33:31  nyef
 * added battery file support
 *
 * Revision 1.20  1999/12/05 00:47:05  nyef
 * cleaned up a bit I missed when adding the new type system
 *
 * Revision 1.19  1999/11/28 17:47:01  nyef
 * fixed type in new system_guess() implementation ('smd', not 'smb')
 *
 * Revision 1.18  1999/11/27 20:09:01  nyef
 * added detect and activation entries for the SG1000
 *
 * Revision 1.17  1999/11/27 18:59:15  nyef
 * built a new version of guess_system() that detects based on filenames
 *
 * Revision 1.16  1999/11/27 18:34:04  nyef
 * added gamegear support to activate_system()
 *
 * Revision 1.15  1999/11/20 05:26:35  nyef
 * redesigned rom loading interface
 * refactored to use standard C file routines in read_romimage()
 * refactored activate_system()
 *
 * Revision 1.14  1999/06/08 01:48:02  nyef
 * added colecovision support (the rom detect is really lame, though)
 *
 * Revision 1.13  1999/04/17 22:24:41  nyef
 * added genesis support.
 *
 * Revision 1.12  1999/02/20 20:12:07  nyef
 * added minimum size checks to SMS/GG detection (fixes small PCE roms)
 * changed most if not all strncmp()s to memcmp()s
 *
 * Revision 1.11  1999/02/14 06:18:19  nyef
 * added detection of FDS images
 *
 * Revision 1.10  1999/01/30 17:53:02  nyef
 * reduced the 0x4000s in the genesis/pce detect to 0x2000 (this doesn't
 * affect the genesis detect much, and fixes the pce detect to catch some
 * roms produced with the MagicKit assembler).
 *
 * Revision 1.9  1999/01/01 06:26:50  nyef
 * added support for Master System.
 *
 * Revision 1.8  1998/12/21 02:58:59  nyef
 * added char * romfilename to save the name of the rom file.
 *
 * Revision 1.7  1998/10/11 20:34:33  nyef
 * added SMS/GG detection to read_romimage()
 * added preliminary GB detection to read_romimage()
 *
 * Revision 1.6  1998/08/02 13:42:40  nyef
 * changed every printf to deb_printf for debug console.
 *
 * Revision 1.5  1998/08/01 01:07:44  nyef
 * fixed stupid mistake in activate_system().
 *
 * Revision 1.4  1998/08/01 01:03:20  nyef
 * moved main() out to the video interfaces. retained some of the logic in
 * activate_system().
 *
 * Revision 1.3  1998/07/12 23:07:07  nyef
 * added a romsize variable and fixed read_romimage() to set it
 *
 * Revision 1.2  1998/07/12 20:52:06  nyef
 * enabled PC-Engine emulation
 *
 * Revision 1.1  1998/07/11 22:19:09  nyef
 * Initial revision
 *
 */
